@@ -1,13 +1,13 @@
 /*
  * axfr.c -- generating AXFR responses.
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "axfr.h"
 #include "dns.h"
@@ -56,7 +56,7 @@ query_axfr(struct nsd *nsd, struct query *query)
 		    || query->axfr_zone->apex != query->domain)
 		{
 			/* No SOA no transfer */
-			RCODE_SET(query->packet, RCODE_REFUSE);
+			RCODE_SET(query->packet, RCODE_NOTAUTH);
 			return QUERY_PROCESSED;
 		}
 
@@ -136,6 +136,7 @@ query_axfr(struct nsd *nsd, struct query *query)
 	}
 
 return_answer:
+	AA_SET(query->packet);
 	ANCOUNT_SET(query->packet, total_added);
 	NSCOUNT_SET(query->packet, 0);
 	ARCOUNT_SET(query->packet, 0);
@@ -156,28 +157,32 @@ return_answer:
 query_state_type
 answer_axfr_ixfr(struct nsd *nsd, struct query *q)
 {
-	acl_options_t *acl;
+	acl_options_t *acl = NULL;
 	/* Is it AXFR? */
 	switch (q->qtype) {
 	case TYPE_AXFR:
 		if (q->tcp) {
-			zone_options_t* zone_opt;
-			zone_opt = zone_options_find(nsd->options, q->qname);
+			zone_options_t* zone_opt = zone_options_find(nsd->options, q->qname);
 			if(!zone_opt ||
 			   acl_check_incoming(zone_opt->provide_xfr, q, &acl)==-1)
 			{
-				char address[128];
-
-				if (addr2ip(q->addr, address, 128)) {
-					DEBUG(DEBUG_XFRD,1, (LOG_INFO,
-						"addr2ip failed"));
-					strlcpy(address, "[unknown]", sizeof(address));
+				if (verbosity > 0) {
+					char address[128];
+					if (addr2ip(q->addr, address, sizeof(address))) {
+						DEBUG(DEBUG_XFRD,1, (LOG_INFO,
+							"addr2ip failed"));
+						strlcpy(address, "[unknown]", sizeof(address));
+					}
+					VERBOSITY(1, (LOG_INFO, "axfr for zone %s from client %s refused, %s",
+						dname_to_string(q->qname, NULL), address, acl?"blocked":"no acl matches"));
 				}
-
-				VERBOSITY(1, (LOG_INFO, "axfr for zone %s from client %s refused, %s", dname_to_string(q->qname, NULL), address, acl?"blocked":"no acl matches"));
 				DEBUG(DEBUG_XFRD,1, (LOG_INFO, "axfr refused, %s",
-						acl?"blocked":"no acl matches"));
-				RCODE_SET(q->packet, RCODE_REFUSE);
+					acl?"blocked":"no acl matches"));
+				if (!zone_opt) {
+					RCODE_SET(q->packet, RCODE_NOTAUTH);
+				} else {
+					RCODE_SET(q->packet, RCODE_REFUSE);
+				}
 				return QUERY_PROCESSED;
 			}
 			DEBUG(DEBUG_XFRD,1, (LOG_INFO, "axfr admitted acl %s %s",
