@@ -1,7 +1,7 @@
 /*
  * nsd-xfer.c -- nsd-xfer(8).
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -14,6 +14,7 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#include <limits.h>
 #include <assert.h>
 #include <errno.h>
 #include <netdb.h>
@@ -45,12 +46,9 @@
  * Exit codes are based on named-xfer for now.  See ns_defs.h in
  * bind8.
  */
-enum nsd_xfer_exit_codes
-{
-	XFER_UPTODATE = 0,
-	XFER_SUCCESS  = 1,
-	XFER_FAIL     = 3
-};
+#define XFER_UPTODATE	0
+#define XFER_SUCCESS	1
+#define XFER_FAIL		3
 
 struct axfr_state
 {
@@ -191,7 +189,7 @@ to_alarm(int ATTR_UNUSED(sig))
  * leading and trailing whitespace and non-zero is returned.
  */
 static int
-read_line(FILE *in, char *line, size_t size)
+tsig_read_line(FILE *in, char *line, size_t size)
 {
 	if (!fgets(line, size, in)) {
 		return 0;
@@ -201,7 +199,6 @@ read_line(FILE *in, char *line, size_t size)
 	}
 }
 
-#ifdef TSIG
 static tsig_key_type *
 read_tsig_key_data(region_type *region, FILE *in,
 	int ATTR_UNUSED(default_family), tsig_algorithm_type** tsig_algo)
@@ -214,7 +211,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	uint8_t data[4000];
 
 	/* server address */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key server address: '%s'",
 		      strerror(errno));
 		return NULL;
@@ -222,7 +219,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	/* server address unused */
 
 	/* tsig keyname */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key name: '%s'", strerror(errno));
 		return NULL;
 	}
@@ -234,7 +231,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	}
 
 	/* tsig algorithm */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key algorithm: '%s'", strerror(errno));
 		return NULL;
 	}
@@ -246,7 +243,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	}
 
 	/* tsig secret */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key data: '%s'\n", strerror(errno));
 		return NULL;
 	}
@@ -294,7 +291,6 @@ read_tsig_key(region_type *region,
 
 	return key;
 }
-#endif /* TSIG */
 
 /*
  * Read SIZE bytes from the socket into BUF.  Keep reading unless an
@@ -704,7 +700,7 @@ init_query(query_type *q,
 	   uint16_t klass,
 	   tsig_record_type *tsig)
 {
-	uint16_t query_id = (uint16_t) random();
+	uint16_t query_id = qid_generate();
 
 	buffer_clear(q->packet);
 
@@ -846,22 +842,21 @@ main(int argc, char *argv[])
 			port = optarg;
 			break;
 		case 's': {
+			uint32_t v;
 			const char *t;
-			state.first_transfer = 0;
-			state.last_serial = strtottl(optarg, &t);
-			if (*t != '\0') {
+ 			state.first_transfer = 0;
+			v = strtoserial(optarg, &t);
+			if (optarg[0] == '\0' || *t != '\0')
+			{
 				error("bad serial '%s'", optarg);
 				exit(XFER_FAIL);
 			}
+			state.last_serial = v;
 			break;
 		}
 		case 'T':
-#ifdef TSIG
 			tsig_key_filename = optarg;
 			break;
-#else
-			log_msg(LOG_ERR, "option -T given but TSIG not enabled");
-#endif /* TSIG */
 		case 'v':
 			++state.verbose;
 			break;
@@ -882,7 +877,6 @@ main(int argc, char *argv[])
 	if (argc == 0 || !zone_filename || !state.zone)
 		usage();
 
-#ifdef TSIG
 	if (tsig_key_filename) {
 		tsig_algorithm_type *tsig_algo = NULL;
 		tsig_key = read_tsig_key(
@@ -899,7 +893,6 @@ main(int argc, char *argv[])
 		tsig_create_record(state.tsig, region);
 		tsig_init_record(state.tsig, tsig_algo, tsig_key);
 	}
-#endif /* TSIG */
 
 	mysigaction.sa_handler = to_alarm;
 	sigfillset(&mysigaction.sa_mask);
@@ -943,7 +936,7 @@ main(int argc, char *argv[])
 		}
 
 		for (res = res0; res; res = res->ai_next) {
-			if (res->ai_addrlen > sizeof(q.addr))
+			if (res->ai_addrlen > (socklen_t)sizeof(q.addr))
 				continue;
 
 			/*
