@@ -1,13 +1,13 @@
 /*
  * query.c -- nsd(8) the resolver.
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -38,6 +38,14 @@
 #include "nsec3.h"
 #include "tsig.h"
 
+/* [Bug #253] Adding unnecessary NS RRset may lead to undesired truncation.
+ * This function determines if the final response packet needs the NS RRset
+ * included. Currently, it will only return negative if QTYPE == DNSKEY|DS.
+ * This way, resolvers won't fallback to TCP unnecessarily when priming
+ * trust anchors.
+ */
+static int answer_needs_ns(struct query  *query);
+
 static int add_rrset(struct query  *query,
 		     answer_type    *answer,
 		     rr_section_type section,
@@ -169,12 +177,10 @@ query_create(region_type *region, uint16_t *compressed_dname_offsets,
 	query->packet = buffer_create(region, QIOBUFSZ);
 	region_add_cleanup(region, query_cleanup, query);
 	query->compressed_dname_offsets_size = compressed_dname_size;
-#ifdef TSIG
 	tsig_create_record(&query->tsig, region);
 	query->tsig_prepare_it = 1;
 	query->tsig_update_it = 1;
 	query->tsig_sign_it = 1;
-#endif /* TSIG */
 	return query;
 }
 
@@ -198,12 +204,10 @@ query_reset(query_type *q, size_t maxlen, int is_tcp)
 	q->reserved_space = 0;
 	buffer_clear(q->packet);
 	edns_init_record(&q->edns);
-#ifdef TSIG
 	tsig_init_record(&q->tsig, NULL, NULL);
 	q->tsig_prepare_it = 1;
 	q->tsig_update_it = 1;
 	q->tsig_sign_it = 1;
-#endif /* TSIG */
 	q->tcp = is_tcp;
 	q->qname = NULL;
 	q->qtype = 0;
@@ -298,21 +302,31 @@ process_query_section(query_type *query)
  * Return NSD_RC_FORMAT on failure, NSD_RC_OK on success.
  */
 static nsd_rc_type
-process_edns(struct query *q)
+process_edns(nsd_type* nsd, struct query *q)
 {
 	if (q->edns.status == EDNS_ERROR) {
+		/* The only error is VERSION not implemented */
 		return NSD_RC_FORMAT;
 	}
+
 	if (q->edns.status == EDNS_OK) {
 		/* Only care about UDP size larger than normal... */
 		if (!q->tcp && q->edns.maxlen > UDP_MAX_MESSAGE_LEN) {
-			if (q->edns.maxlen < EDNS_MAX_MESSAGE_LEN) {
+			size_t edns_size;
+#if defined(INET6)
+			if (q->addr.ss_family == AF_INET6) {
+				edns_size = nsd->ipv6_edns_size;
+			} else
+#endif
+			edns_size = nsd->ipv4_edns_size;
+
+			if (q->edns.maxlen < edns_size) {
 				q->maxlen = q->edns.maxlen;
 			} else {
-				q->maxlen = EDNS_MAX_MESSAGE_LEN;
+				q->maxlen = edns_size;
 			}
 
-#if defined(INET6) && !defined(IPV6_USE_MIN_MTU)
+#if defined(INET6) && !defined(IPV6_USE_MIN_MTU) && !defined(IPV6_MTU)
 			/*
 			 * Use IPv6 minimum MTU to avoid sending
 			 * packets that are too large for some links.
@@ -339,7 +353,6 @@ process_edns(struct query *q)
  * Processes TSIG.
  * Sets error when tsig does not verify on the query.
  */
-#ifdef TSIG
 static nsd_rc_type
 process_tsig(struct query* q)
 {
@@ -364,7 +377,6 @@ process_tsig(struct query* q)
 	}
 	return NSD_RC_OK;
 }
-#endif /* TSIG */
 
 /*
  * Check notify acl and forward to xfrd (or return an error).
@@ -387,7 +399,6 @@ answer_notify(struct nsd* nsd, struct query *query)
 	if(!nsd->this_child) /* we are in debug mode or something */
 		return query_error(query, NSD_RC_SERVFAIL);
 
-#ifdef TSIG
 	if(!tsig_find_rr(&query->tsig, query->packet)) {
 		DEBUG(DEBUG_XFRD,2, (LOG_ERR, "bad tsig RR format"));
 		return query_error(query, NSD_RC_FORMAT);
@@ -395,7 +406,6 @@ answer_notify(struct nsd* nsd, struct query *query)
 	rc = process_tsig(query);
 	if(rc != NSD_RC_OK)
 		return query_error(query, rc);
-#endif /* TSIG */
 
 	/* check if it passes acl */
 	if((acl_num = acl_check_incoming(zone_opt->allow_notify, query,
@@ -443,10 +453,21 @@ answer_notify(struct nsd* nsd, struct query *query)
 		/* tsig is added in add_additional later (if needed) */
 		return QUERY_PROCESSED;
 	}
-	VERBOSITY(1, (LOG_INFO, "got notify for zone: %s; Refused by acl: %s %s",
+
+	if (verbosity > 1) {
+		char address[128];
+		if (addr2ip(query->addr, address, sizeof(address))) {
+			DEBUG(DEBUG_XFRD,1, (LOG_INFO, "addr2ip failed"));
+			strlcpy(address, "[unknown]", sizeof(address));
+		}
+
+		VERBOSITY(1, (LOG_INFO, "notify for zone %s from client %s refused, %s%s",
 			dname_to_string(query->qname, NULL),
+			address,
 			why?why->key_name:"no acl matches",
 			why?why->ip_address_spec:"."));
+	}
+
 	return query_error(query, NSD_RC_REFUSE);
 }
 
@@ -605,6 +626,15 @@ add_additional_rrsets(struct query *query, answer_type *answer,
 	}
 }
 
+static int
+answer_needs_ns(struct query* query)
+{
+	assert(query);
+	/* Currently, only troublesome for DNSKEY and DS,
+         * cuz their RRSETs are quite large. */
+	return (query->qtype != TYPE_DNSKEY && query->qtype != TYPE_DS);
+}
+
 static int
 add_rrset(struct query   *query,
 	  answer_type    *answer,
@@ -690,11 +720,11 @@ query_synthesize_cname(struct query* q, struct answer* answer, const dname_type*
 
 	/* allocate dest part */
 	lastparent = to_closest_encloser;
-	for(i=0; i < to_name->label_count - domain_dname(to_closest_encloser)->label_count; 
+	for(i=0; i < to_name->label_count - domain_dname(to_closest_encloser)->label_count;
 		i++)
 	{
 		domain_type* newdom = query_get_tempdomain(q);
-		if(!newdom) 
+		if(!newdom)
 			return 0;
 		newdom->is_existing = 0;
 		newdom->parent = lastparent;
@@ -744,7 +774,12 @@ answer_delegation(query_type *query, answer_type *answer)
 	assert(query->delegation_domain);
 	assert(query->delegation_rrset);
 
-	AA_CLR(query->packet);
+	if (query->cname_count == 0) {
+		AA_CLR(query->packet);
+	} else {
+		AA_SET(query->packet);
+	}
+
 	add_rrset(query,
 		  answer,
 		  AUTHORITY_SECTION,
@@ -894,6 +929,9 @@ answer_domain(struct nsd* nsd, struct query *q, answer_type *answer,
 					     domain_dname(closest_match));
 			q->zone = origzone;
 		}
+		/* example 6.2.7 shows no NS-set from zone in auth (RFC1034) */
+		q->domain = domain;
+		return;
 	} else {
 		answer_nodata(q, answer, original);
 		return;
@@ -901,8 +939,8 @@ answer_domain(struct nsd* nsd, struct query *q, answer_type *answer,
 
 	q->domain = domain;
 
-	if (q->qclass != CLASS_ANY && q->zone->ns_rrset) {
-		add_rrset(q, answer, AUTHORITY_SECTION, q->zone->apex,
+	if (q->qclass != CLASS_ANY && q->zone->ns_rrset && answer_needs_ns(q)) {
+		add_rrset(q, answer, OPTIONAL_AUTHORITY_SECTION, q->zone->apex,
 			  q->zone->ns_rrset);
 	}
 }
@@ -1003,12 +1041,13 @@ answer_authoritative(struct nsd   *nsd,
 		match->rrsets = wildcard_child->rrsets;
 		match->is_existing = wildcard_child->is_existing;
 #ifdef NSEC3
-		match->nsec3_is_exact = wildcard_child->nsec3_is_exact;
 		match->nsec3_cover = wildcard_child->nsec3_cover;
+#ifdef FULL_PREHASH
+		match->nsec3_is_exact = wildcard_child->nsec3_is_exact;
 		match->nsec3_wcard_child_cover = wildcard_child->nsec3_wcard_child_cover;
 		match->nsec3_ds_parent_is_exact = wildcard_child->nsec3_ds_parent_is_exact;
 		match->nsec3_ds_parent_cover = wildcard_child->nsec3_ds_parent_cover;
-
+#endif
 		if (q->edns.dnssec_ok && q->zone->nsec3_soa_rr) {
 			/* Only add nsec3 wildcard data when do bit is set */
 			nsec3_answer_wildcard(q, answer, wildcard_child, nsd->db, qname);
@@ -1170,11 +1209,17 @@ answer_query(struct nsd *nsd, struct query *q)
 	answer_lookup_zone(nsd, q, &answer, 0, exact, closest_match,
 		closest_encloser, q->qname);
 
+#ifdef USE_ZONE_STATS
+	if (q->zone) {
+		ZTATUP2(q->zone, opcode, q->opcode);
+		ZTATUP2(q->zone, qtype, q->qtype);
+		ZTATUP2(q->zone, opcode, q->qclass);
+	}
+#endif
+
 	offset = dname_label_offsets(q->qname)[domain_dname(closest_encloser)->label_count - 1] + QHEADERSZ;
 	query_add_compression_domain(q, closest_encloser, offset);
-
 	encode_answer(q, &answer);
-
 	query_clear_compression_tables(q);
 }
 
@@ -1193,9 +1238,7 @@ query_prepare_response(query_type *q)
 	 * Reserve space for the EDNS records if required.
 	 */
 	q->reserved_space = edns_reserved_space(&q->edns);
-#ifdef TSIG
 	q->reserved_space += tsig_reserved_space(&q->tsig);
-#endif /* TSIG */
 
 	/* Update the flags.  */
 	flags = FLAGS(q->packet);
@@ -1228,7 +1271,7 @@ query_process(query_type *q, nsd_type *nsd)
 		return QUERY_DISCARDED;
 	}
 
-	if(!process_query_section(q)) {
+	if (RCODE(q->packet) != RCODE_OK || !process_query_section(q)) {
 		return query_formerr(q);
 	}
 
@@ -1246,10 +1289,11 @@ query_process(query_type *q, nsd_type *nsd)
 	}
 
 	/* Dont bother to answer more than one question at once... */
-	if (QDCOUNT(q->packet) != 1 || TC(q->packet)) {
+	if (QDCOUNT(q->packet) != 1) {
 		FLAGS_SET(q->packet, 0);
 		return query_formerr(q);
 	}
+	/* Ignore settings of flags */
 
 	/* Dont allow any records in the answer or authority section...
 	   except for IXFR queries. */
@@ -1265,7 +1309,6 @@ query_process(query_type *q, nsd_type *nsd)
 	}
 
 	arcount = ARCOUNT(q->packet);
-#ifdef TSIG
 	if (arcount > 0) {
 		/* see if tsig is before edns record */
 		if (!tsig_parse_rr(&q->tsig, q->packet))
@@ -1273,12 +1316,10 @@ query_process(query_type *q, nsd_type *nsd)
 		if(q->tsig.status != TSIG_NOT_PRESENT)
 			--arcount;
 	}
-#endif /* TSIG */
 	if (arcount > 0) {
 		if (edns_parse_record(&q->edns, q->packet))
 			--arcount;
 	}
-#ifdef TSIG
 	if (arcount > 0 && q->tsig.status == TSIG_NOT_PRESENT) {
 		/* see if tsig is after the edns record */
 		if (!tsig_parse_rr(&q->tsig, q->packet))
@@ -1286,7 +1327,6 @@ query_process(query_type *q, nsd_type *nsd)
 		if(q->tsig.status != TSIG_NOT_PRESENT)
 			--arcount;
 	}
-#endif /* TSIG */
 	if (arcount > 0) {
 		return query_formerr(q);
 	}
@@ -1301,15 +1341,17 @@ query_process(query_type *q, nsd_type *nsd)
 	/* Remove trailing garbage.  */
 	buffer_set_limit(q->packet, buffer_position(q->packet));
 
-#ifdef TSIG
 	rc = process_tsig(q);
 	if (rc != NSD_RC_OK) {
 		return query_error(q, rc);
 	}
-#endif /* TSIG */
-	rc = process_edns(q);
+	rc = process_edns(nsd, q);
 	if (rc != NSD_RC_OK) {
-		return query_error(q, rc);
+		/* We should not return FORMERR, but BADVERS (=16).
+		 * BADVERS is created with Ext. RCODE, followed by RCODE.
+		 * Ext. RCODE is set to 1, RCODE must be 0 (getting 0x10 = 16).
+		 * Thus RCODE = NOERROR = NSD_RC_OK. */
+		return query_error(q, NSD_RC_OK);
 	}
 
 	query_prepare_response(q);
@@ -1341,13 +1383,16 @@ query_add_optional(query_type *q, nsd_type *nsd)
 		edns = &nsd->edns_ipv6;
 	}
 #endif
+	if (RCODE(q->packet) == RCODE_FORMAT) {
+		return;
+	}
 	switch (q->edns.status) {
 	case EDNS_NOT_PRESENT:
 		break;
 	case EDNS_OK:
+		if (q->edns.dnssec_ok)	edns->ok[7] = 0x80;
+		else			edns->ok[7] = 0x00;
 		buffer_write(q->packet, edns->ok, OPT_LEN);
-		/* check if nsid data should be written */
-#ifdef NSID
 		if (nsd->nsid_len > 0 && q->edns.nsid == 1 &&
 				!query_overflow_nsid(q, nsd->nsid_len)) {
 			/* rdata length */
@@ -1360,23 +1405,29 @@ query_add_optional(query_type *q, nsd_type *nsd)
 			/* fill with NULLs */
 			buffer_write(q->packet, edns->rdata_none, OPT_RDATA);
 		}
-#else
-		buffer_write(q->packet, edns->rdata_none, OPT_RDATA);
-#endif /* NSID */
-
 		ARCOUNT_SET(q->packet, ARCOUNT(q->packet) + 1);
-
 		STATUP(nsd, edns);
+#ifdef USE_ZONE_STATS
+		if (q->zone) {
+			ZTATUP(q->zone, edns);
+		}
+#endif
 		break;
 	case EDNS_ERROR:
+		if (q->edns.dnssec_ok)	edns->error[7] = 0x80;
+		else			edns->error[7] = 0x00;
 		buffer_write(q->packet, edns->error, OPT_LEN);
+		buffer_write(q->packet, edns->rdata_none, OPT_RDATA);
 		ARCOUNT_SET(q->packet, ARCOUNT(q->packet) + 1);
-
 		STATUP(nsd, ednserr);
+#ifdef USE_ZONE_STATS
+		if (q->zone) {
+			ZTATUP(q->zone, ednserr);
+		}
+#endif
 		break;
 	}
 
-#ifdef TSIG
 	if (q->tsig.status != TSIG_NOT_PRESENT) {
 		if (q->tsig.status == TSIG_ERROR ||
 			q->tsig.error_code != TSIG_ERROR_NOERROR) {
@@ -1397,5 +1448,4 @@ query_add_optional(query_type *q, nsd_type *nsd)
 			}
 		}
 	}
-#endif /* TSIG */
 }
