@@ -1,7 +1,7 @@
 /*
  * rdata.h -- RDATA conversion functions.
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -21,7 +21,7 @@ extern lookup_table_type dns_certificate_types[];
 extern lookup_table_type dns_algorithms[];
 
 int rdata_atom_to_string(buffer_type *output, rdata_zoneformat_type type,
-			 rdata_atom_type rdata);
+			 rdata_atom_type rdata, rr_type *rr);
 
 /*
  * Split the wireformat RDATA into an array of rdata atoms. Domain
@@ -51,4 +51,10 @@ int rdata_atoms_to_unknown_string(buffer_type *out,
 				  size_t rdata_count,
 				  rdata_atom_type *rdatas);
 
+/* print rdata to a text string (as for a zone file) returns 0
+  on a failure (bufpos is reset to original position).
+  returns 1 on success, bufpos is moved. */
+int print_rdata(buffer_type *output, rrtype_descriptor_type *descriptor,
+            rr_type *record);
+
 #endif /* _DNS_H_ */
