@@ -1,13 +1,13 @@
 /*
  * server.c -- nsd(8) network input/output
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -29,6 +29,9 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <netdb.h>
+#ifndef SHUT_WR
+#define SHUT_WR 1
+#endif
 
 #include "axfr.h"
 #include "namedb.h"
@@ -60,6 +63,9 @@ struct tcp_accept_handler_data {
 	netio_handler_type *tcp_accept_handlers;
 };
 
+int slowaccept;
+struct timespec slowaccept_timeout;
+
 /*
  * Data for the TCP connection handlers.
  *
@@ -114,6 +120,11 @@ struct tcp_handler_data
 	 * to specify the packet length on a TCP connection.
 	 */
 	size_t           bytes_transmitted;
+
+	/*
+	 * The number of queries handled by this specific TCP connection.
+	 */
+	int					query_count;
 };
 
 /*
@@ -160,7 +171,9 @@ static void handle_tcp_writing(netio_type *netio,
 static void send_children_quit(struct nsd* nsd);
 
 /* set childrens flags to send NSD_STATS to them */
+#ifdef BIND8_STATS
 static void set_children_stats(struct nsd* nsd);
+#endif /* BIND8_STATS */
 
 /*
  * Change the event types the HANDLERS are interested in to
@@ -192,7 +205,7 @@ delete_child_pid(struct nsd *nsd, pid_t pid)
 		if (nsd->children[i].pid == pid) {
 			nsd->children[i].pid = 0;
 			if(!nsd->children[i].need_to_exit) {
-				if(nsd->children[i].child_fd > 0)
+				if(nsd->children[i].child_fd != -1)
 					close(nsd->children[i].child_fd);
 				nsd->children[i].child_fd = -1;
 				if(nsd->children[i].handler)
@@ -218,7 +231,7 @@ restart_child_servers(struct nsd *nsd, region_type* region, netio_type* netio,
 	/* Fork the child processes... */
 	for (i = 0; i < nsd->child_count; ++i) {
 		if (nsd->children[i].pid <= 0) {
-			if (nsd->children[i].child_fd > 0)
+			if (nsd->children[i].child_fd != -1)
 				close(nsd->children[i].child_fd);
 			if (socketpair(AF_UNIX, SOCK_STREAM, 0, sv) == -1) {
 				log_msg(LOG_ERR, "socketpair: %s",
@@ -315,6 +328,12 @@ initialize_dname_compression_tables(struct nsd *nsd)
 	size_t needed = domain_table_count(nsd->db->domains) + 1;
 	needed += EXTRA_DOMAIN_NUMBERS;
 	if(compression_table_capacity < needed) {
+		if(compressed_dname_offsets) {
+			region_remove_cleanup(nsd->db->region,
+				cleanup_dname_compression_tables,
+				compressed_dname_offsets);
+			free(compressed_dname_offsets);
+		}
 		compressed_dname_offsets = (uint16_t *) xalloc(
 			needed * sizeof(uint16_t));
 		region_add_cleanup(nsd->db->region, cleanup_dname_compression_tables,
@@ -326,39 +345,15 @@ initialize_dname_compression_tables(struct nsd *nsd)
 	compressed_dname_offsets[0] = QHEADERSZ; /* The original query name */
 }
 
-static int
-file_inside_chroot(const char* fname, const char* chr)
-{
-#ifdef NDEBUG
-	assert(chr);
-#endif /* NDEBUG */
-	/* logfile and chroot the same? */
-	if (fname && fname[0] && chr[0] && !strncmp(fname, chr, strlen(chr)))
-		return 2; /* strip chroot, file rotation ok */
-	else if (fname && fname[0] != '/')
-		return 1; /* don't strip, file rotation ok */
-	return 0; /* don't strip, don't try file rotation */
-}
-
-static void
-pid_unlink(const char* pidfile)
-{
-	if (pidfile && unlink(pidfile) == -1)
-		log_msg(LOG_ERR, "failed to unlink pidfile %s: %s",
-			pidfile, strerror(errno));
-}
-
 /*
  * Initialize the server, create and bind the sockets.
- * Drop the privileges and chroot if requested.
  *
  */
 int
 server_init(struct nsd *nsd)
 {
-	FILE* dbfd;
 	size_t i;
-#if defined(SO_REUSEADDR) || (defined(INET6) && (defined(IPV6_V6ONLY) || defined(IPV6_USE_MIN_MTU)))
+#if defined(SO_REUSEADDR) || (defined(INET6) && (defined(IPV6_V6ONLY) || defined(IPV6_USE_MIN_MTU) || defined(IPV6_MTU)))
 	int on = 1;
 #endif
 
@@ -411,7 +406,44 @@ server_init(struct nsd *nsd)
 					strerror(errno));
 				return -1;
 			}
+# elif defined(IPV6_MTU)
+			/*
+			 * On Linux, PMTUD is disabled by default for datagrams
+			 * so set the MTU equal to the MIN MTU to get the same.
+			 */
+			on = IPV6_MIN_MTU;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IPV6, IPV6_MTU, 
+				&on, sizeof(on)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IPV6_MTU, ...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+			on = 1;
+# endif
+		}
 #endif
+#if defined(AF_INET)
+		if (nsd->udp[i].addr->ai_family == AF_INET) {
+#  if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_DONT)
+			int action = IP_PMTUDISC_DONT;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IP, 
+				IP_MTU_DISCOVER, &action, sizeof(action)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IP_MTU_DISCOVER, IP_PMTUDISC_DONT...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+#  elif defined(IP_DONTFRAG)
+			int off = 0;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IP, IP_DONTFRAG,
+				&off, sizeof(off)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IP_DONTFRAG, ...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+#  endif
 		}
 #endif
 		/* set it nonblocking */
@@ -482,78 +514,29 @@ server_init(struct nsd *nsd)
 		}
 	}
 
-#ifdef HAVE_CHROOT
-	/* Chroot */
-	if (nsd->chrootdir && strlen(nsd->chrootdir)) {
-		int l = strlen(nsd->chrootdir);
-		int ret = 0;
-
-		while (l>0 && nsd->chrootdir[l-1] == '/')
-			--l;
-
-		/* logfile */
-		ret = file_inside_chroot(nsd->log_filename, nsd->chrootdir);
-		if (ret)
-		{
-			nsd->file_rotation_ok = 1;
-			if (ret == 2) /* also strip chroot */
-				nsd->log_filename += l;
-		}
-
-		nsd->dbfile += l;
-		nsd->pidfile += l;
-		nsd->options->xfrdfile += l;
-		nsd->options->difffile += l;
-
-		if (chroot(nsd->chrootdir)) {
-			log_msg(LOG_ERR, "unable to chroot: %s", strerror(errno));
-			return -1;
-		}
-		DEBUG(DEBUG_IPC,1, (LOG_INFO, "changed root directory to %s",
-			nsd->chrootdir));
-	}
-	else
-#endif
-		nsd->file_rotation_ok = 1;
-
-	DEBUG(DEBUG_IPC,1, (LOG_INFO, "file rotation on %s %sabled",
-		nsd->log_filename, nsd->file_rotation_ok?"en":"dis"));
-
-	/* Check if nsd->dbfile exists */
-	if ((dbfd = fopen(nsd->dbfile, "r")) == NULL) {
-		log_msg(LOG_ERR, "unable to open %s for reading: %s", nsd->dbfile, strerror(errno));
-		return -1;
-	}
-	fclose(dbfd);
-
-	/* Write pidfile */
-	if (writepid(nsd) == -1) {
-		log_msg(LOG_ERR, "cannot overwrite the pidfile %s: %s",
-			nsd->pidfile, strerror(errno));
-	}
-
-	/* Drop the permissions */
-	if (setgid(nsd->gid) != 0 || setuid(nsd->uid) !=0) {
-		log_msg(LOG_ERR, "unable to drop user privileges: %s",
-			strerror(errno));
-		pid_unlink(nsd->pidfile);
-		return -1;
-	}
+	return 0;
+}
 
+/*
+ * Prepare the server for take off.
+ *
+ */
+int
+server_prepare(struct nsd *nsd)
+{
 	/* Open the database... */
 	if ((nsd->db = namedb_open(nsd->dbfile, nsd->options, nsd->child_count)) == NULL) {
 		log_msg(LOG_ERR, "unable to open the database %s: %s",
 			nsd->dbfile, strerror(errno));
-		pid_unlink(nsd->pidfile);
 		return -1;
 	}
 
 	/* Read diff file */
-	if(!diff_read_file(nsd->db, nsd->options, NULL, nsd->child_count))
-	{
+	if(!diff_read_file(nsd->db, nsd->options, NULL, nsd->child_count)) {
 		log_msg(LOG_ERR, "The diff file contains errors. Will continue "
 						 "without it");
 	}
+
 #ifdef NSEC3
 	prehash(nsd->db, 0);
 #endif
@@ -596,7 +579,7 @@ close_all_sockets(struct nsd_socket sockets[], size_t n)
 	for (i = 0; i < n; ++i) {
 		if (sockets[i].s != -1) {
 			close(sockets[i].s);
-			free(sockets[i].addr);
+			freeaddrinfo(sockets[i].addr);
 			sockets[i].s = -1;
 		}
 	}
@@ -615,7 +598,7 @@ server_shutdown(struct nsd *nsd)
 	close_all_sockets(nsd->udp, nsd->ifs);
 	close_all_sockets(nsd->tcp, nsd->ifs);
 	/* CHILD: close command channel to parent */
-	if(nsd->this_child && nsd->this_child->parent_fd > 0)
+	if(nsd->this_child && nsd->this_child->parent_fd != -1)
 	{
 		close(nsd->this_child->parent_fd);
 		nsd->this_child->parent_fd = -1;
@@ -624,7 +607,7 @@ server_shutdown(struct nsd *nsd)
 	if(!nsd->this_child)
 	{
 		for(i=0; i < nsd->child_count; ++i)
-			if(nsd->children[i].child_fd > 0)
+			if(nsd->children[i].child_fd != -1)
 			{
 				close(nsd->children[i].child_fd);
 				nsd->children[i].child_fd = -1;
@@ -706,7 +689,7 @@ block_read(struct nsd* nsd, int s, void* p, ssize_t sz, int timeout)
 				/* blocking read */
 				continue;
 			if(errno == EINTR) {
-				if(nsd->signal_hint_quit || nsd->signal_hint_shutdown)
+				if(nsd && (nsd->signal_hint_quit || nsd->signal_hint_shutdown))
 					return -1;
 				/* other signals can be handled later */
 				continue;
@@ -724,7 +707,7 @@ block_read(struct nsd* nsd, int s, void* p, ssize_t sz, int timeout)
 				/* blocking read */
 				continue;
 			if(errno == EINTR) {
-				if(nsd->signal_hint_quit || nsd->signal_hint_shutdown)
+				if(nsd && (nsd->signal_hint_quit || nsd->signal_hint_shutdown))
 					return -1;
 				/* other signals can be handled later */
 				continue;
@@ -779,7 +762,9 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		region_log_stats(nsd->db->region);
 #endif /* NDEBUG */
 #ifdef NSEC3
+#ifdef FULL_PREHASH
 	prehash(nsd->db, 1);
+#endif /* FULL_PREHASH */
 #endif /* NSEC3 */
 
 	initialize_dname_compression_tables(nsd);
@@ -800,6 +785,16 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		exit(1);
 	}
 
+	/* if the parent has quit, we must quit too, poll the fd for cmds */
+	if(block_read(nsd, cmdsocket, &cmd, sizeof(cmd), 0) == sizeof(cmd)) {
+		DEBUG(DEBUG_IPC,1, (LOG_INFO, "reload: ipc command from main %d", cmd));
+		if(cmd == NSD_QUIT) {
+			DEBUG(DEBUG_IPC,1, (LOG_INFO, "reload: quit to follow nsd"));
+			send_children_quit(nsd);
+			exit(0);
+		}
+	}
+
 	/* Overwrite pid before closing old parent, to avoid race condition:
 	 * - parent process already closed
 	 * - pidfile still contains old_pid
@@ -831,6 +826,12 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 			strerror(errno));
 	}
 	DEBUG(DEBUG_IPC,1, (LOG_INFO, "reload: ipc reply main %d %d", ret, cmd));
+	if(cmd == NSD_QUIT) {
+		/* small race condition possible here, parent got quit cmd. */
+		send_children_quit(nsd);
+		unlinkpid(nsd->pidfile);
+		exit(1);
+	}
 	assert(ret==-1 || ret == 0 || cmd == NSD_RELOAD);
 
 	/* inform xfrd of new SOAs */
@@ -1113,7 +1114,7 @@ server_main(struct nsd *nsd)
 			break;
 		case NSD_QUIT_SYNC:
 			/* synchronisation of xfrd, parent and reload */
-			if(!nsd->quit_sync_done && reload_listener.fd > 0) {
+			if(!nsd->quit_sync_done && reload_listener.fd != -1) {
 				sig_atomic_t cmd = NSD_RELOAD;
 				/* stop xfrd ipc writes in progress */
 				DEBUG(DEBUG_IPC,1, (LOG_INFO,
@@ -1130,7 +1131,7 @@ server_main(struct nsd *nsd)
 			break;
 		case NSD_QUIT:
 			/* silent shutdown during reload */
-			if(reload_listener.fd > 0) {
+			if(reload_listener.fd != -1) {
 				/* acknowledge the quit, to sync reload that we will really quit now */
 				sig_atomic_t cmd = NSD_RELOAD;
 				DEBUG(DEBUG_IPC,1, (LOG_INFO, "main: ipc ack reload"));
@@ -1143,8 +1144,8 @@ server_main(struct nsd *nsd)
 			/* only quit children after xfrd has acked */
 			send_children_quit(nsd);
 
+			namedb_fd_close(nsd->db);
 			region_destroy(server_region);
-			namedb_close(nsd->db);
 			server_shutdown(nsd);
 
 			/* ENOTREACH */
@@ -1177,11 +1178,20 @@ server_main(struct nsd *nsd)
 	close(fd);
 
 	/* Unlink it if possible... */
-	pid_unlink(nsd->pidfile);
+	unlinkpid(nsd->pidfile);
 
-	if(reload_listener.fd > 0)
+	if(reload_listener.fd != -1) {
+		sig_atomic_t cmd = NSD_QUIT;
+		DEBUG(DEBUG_IPC,1, (LOG_INFO,
+			"main: ipc send quit to reload-process"));
+		if(!write_socket(reload_listener.fd, &cmd, sizeof(cmd))) {
+			log_msg(LOG_ERR, "server_main: could not send quit to reload: %s",
+				strerror(errno));
+		}
+		fsync(reload_listener.fd);
 		close(reload_listener.fd);
-	if(xfrd_listener.fd > 0) {
+	}
+	if(xfrd_listener.fd != -1) {
 		/* complete quit, stop xfrd */
 		sig_atomic_t cmd = NSD_QUIT;
 		DEBUG(DEBUG_IPC,1, (LOG_INFO,
@@ -1192,9 +1202,10 @@ server_main(struct nsd *nsd)
 		}
 		fsync(xfrd_listener.fd);
 		close(xfrd_listener.fd);
+		(void)kill(xfrd_pid, SIGTERM);
 	}
 
-	namedb_close(nsd->db);
+	namedb_fd_close(nsd->db);
 	region_destroy(server_region);
 	server_shutdown(nsd);
 }
@@ -1296,7 +1307,7 @@ server_child(struct nsd *nsd)
 			handler->fd = nsd->tcp[i].s;
 			handler->timeout = NULL;
 			handler->user_data = data;
-			handler->event_types = NETIO_EVENT_READ;
+			handler->event_types = NETIO_EVENT_READ | NETIO_EVENT_ACCEPT;
 			handler->event_handler = handle_tcp_accept;
 			netio_add_handler(netio, handler);
 		}
@@ -1319,7 +1330,7 @@ server_child(struct nsd *nsd)
 		}
 		else if (mode == NSD_REAP_CHILDREN) {
 			/* got signal, notify parent. parent reaps terminated children. */
-			if (nsd->this_child->parent_fd > 0) {
+			if (nsd->this_child->parent_fd != -1) {
 				sig_atomic_t parent_notify = NSD_REAP_CHILDREN;
 				if (write(nsd->this_child->parent_fd,
 				    &parent_notify,
@@ -1353,7 +1364,7 @@ server_child(struct nsd *nsd)
 	bind8_stats(nsd);
 #endif /* BIND8_STATS */
 
-	namedb_close(nsd->db);
+	namedb_fd_close(nsd->db);
 	region_destroy(server_region);
 	server_shutdown(nsd);
 }
@@ -1374,11 +1385,13 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 	}
 
 	/* Account... */
+#ifdef BIND8_STATS
 	if (data->socket->addr->ai_family == AF_INET) {
 		STATUP(data->nsd, qudp);
 	} else if (data->socket->addr->ai_family == AF_INET6) {
 		STATUP(data->nsd, qudp6);
 	}
+#endif
 
 	/* Initialize the query... */
 	query_reset(q, UDP_MAX_MESSAGE_LEN, 0);
@@ -1393,6 +1406,7 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 		if (errno != EAGAIN && errno != EINTR) {
 			log_msg(LOG_ERR, "recvfrom failed: %s", strerror(errno));
 			STATUP(data->nsd, rxerr);
+			/* No zone statup */
 		}
 	} else {
 		buffer_skip(q->packet, received);
@@ -1400,8 +1414,25 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 
 		/* Process and answer the query... */
 		if (server_process_query(data->nsd, q) != QUERY_DISCARDED) {
-			if (RCODE(q->packet) == RCODE_OK && !AA(q->packet))
+#ifdef BIND8_STATS
+			if (RCODE(q->packet) == RCODE_OK && !AA(q->packet)) {
 				STATUP(data->nsd, nona);
+# ifdef USE_ZONE_STATS
+				if (q->zone)
+					ZTATUP(q->zone, nona);
+# endif
+			}
+
+# ifdef USE_ZONE_STATS
+			if (q->zone) {
+			  if (data->socket->addr->ai_family == AF_INET) {
+				ZTATUP(q->zone, qudp);
+			  } else if (data->socket->addr->ai_family == AF_INET6) {
+				ZTATUP(q->zone, qudp6);
+			  }
+			}
+# endif
+#endif
 
 			/* Add EDNS0 and TSIG info if necessary.  */
 			query_add_optional(q, data->nsd);
@@ -1417,18 +1448,39 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 			if (sent == -1) {
 				log_msg(LOG_ERR, "sendto failed: %s", strerror(errno));
 				STATUP(data->nsd, txerr);
+
+#ifdef USE_ZONE_STATS
+				if (q->zone)
+					ZTATUP(q->zone, txerr);
+#endif
 			} else if ((size_t) sent != buffer_remaining(q->packet)) {
 				log_msg(LOG_ERR, "sent %d in place of %d bytes", sent, (int) buffer_remaining(q->packet));
-			} else {
 #ifdef BIND8_STATS
+			} else {
 				/* Account the rcode & TC... */
 				STATUP2(data->nsd, rcode, RCODE(q->packet));
-				if (TC(q->packet))
+# ifdef USE_ZONE_STATS
+				if (q->zone)
+					ZTATUP2(q->zone, rcode, RCODE(q->packet));
+# endif
+				if (TC(q->packet)) {
 					STATUP(data->nsd, truncated);
+# ifdef USE_ZONE_STATS
+					if (q->zone)
+						ZTATUP(q->zone, truncated);
+# endif
+				}
 #endif /* BIND8_STATS */
 			}
+#ifdef BIND8_STATS
 		} else {
 			STATUP(data->nsd, dropped);
+# ifdef USE_ZONE_STATS
+			if (q->zone) {
+				ZTATUP(q->zone, dropped);
+			}
+# endif
+#endif
 		}
 	}
 }
@@ -1441,6 +1493,7 @@ cleanup_tcp_handler(netio_type *netio, netio_handler_type *handler)
 		= (struct tcp_handler_data *) handler->user_data;
 	netio_remove_handler(netio, handler);
 	close(handler->fd);
+	slowaccept = 0;
 
 	/*
 	 * Enable the TCP accept handlers when the current number of
@@ -1473,6 +1526,13 @@ handle_tcp_reading(netio_type *netio,
 		return;
 	}
 
+	if (data->nsd->tcp_query_count > 0 &&
+		data->query_count >= data->nsd->tcp_query_count) {
+		/* No more queries allowed on this tcp connection.  */
+		cleanup_tcp_handler(netio, handler);
+		return;
+	}
+
 	assert(event_types & NETIO_EVENT_READ);
 
 	if (data->bytes_transmitted == 0) {
@@ -1584,33 +1644,64 @@ handle_tcp_reading(netio_type *netio,
 	assert(buffer_position(data->query->packet) == data->query->tcplen);
 
 	/* Account... */
-#ifndef INET6
-        STATUP(data->nsd, ctcp);
-#else
+#ifdef BIND8_STATS
+# ifndef INET6
+	STATUP(data->nsd, ctcp);
+# else
 	if (data->query->addr.ss_family == AF_INET) {
 		STATUP(data->nsd, ctcp);
 	} else if (data->query->addr.ss_family == AF_INET6) {
 		STATUP(data->nsd, ctcp6);
 	}
-#endif
+# endif
+#endif /* BIND8_STATS */
 
 	/* We have a complete query, process it.  */
 
+	/* tcp-query-count: handle query counter ++ */
+	data->query_count++;
+
 	buffer_flip(data->query->packet);
 	data->query_state = server_process_query(data->nsd, data->query);
 	if (data->query_state == QUERY_DISCARDED) {
 		/* Drop the packet and the entire connection... */
 		STATUP(data->nsd, dropped);
+#if defined(BIND8_STATS) && defined(USE_ZONE_STATS)
+		if (data->query->zone) {
+			ZTATUP(data->query->zone, dropped);
+		}
+#endif
 		cleanup_tcp_handler(netio, handler);
 		return;
 	}
 
+#ifdef BIND8_STATS
 	if (RCODE(data->query->packet) == RCODE_OK
 	    && !AA(data->query->packet))
 	{
 		STATUP(data->nsd, nona);
+# ifdef USE_ZONE_STATS
+		if (data->query->zone)
+			ZTATUP(data->query->zone, nona);
+# endif
 	}
 
+# ifdef USE_ZONE_STATS
+	if (data->query->zone) {
+#  ifndef INET6
+		ZTATUP(data->query->zone, ctcp);
+#  else
+	if (data->query->addr.ss_family == AF_INET) {
+		ZTATUP(data->query->zone, ctcp);
+	} else if (data->query->addr.ss_family == AF_INET6) {
+		ZTATUP(data->query->zone, ctcp6);
+	}
+#  endif
+	}
+# endif /* USE_ZONE_STATS */
+
+#endif /* BIND8_STATS */
+
 	query_add_optional(data->query, data->nsd);
 
 	/* Switch to the tcp write handler.  */
@@ -1618,7 +1709,7 @@ handle_tcp_reading(netio_type *netio,
 	data->query->tcplen = buffer_remaining(data->query->packet);
 	data->bytes_transmitted = 0;
 
-	handler->timeout->tv_sec = TCP_TIMEOUT;
+	handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	handler->timeout->tv_nsec = 0L;
 	timespec_add(handler->timeout, netio_current_time(netio));
 
@@ -1661,6 +1752,9 @@ handle_tcp_writing(netio_type *netio,
 #ifdef ECONNRESET
 				if(verbosity >= 2 || errno != ECONNRESET)
 #endif /* ECONNRESET */
+#ifdef EPIPE
+					if(verbosity >= 2 || errno != EPIPE)
+#endif /* EPIPE 'broken pipe' */
 				log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 				cleanup_tcp_handler(netio, handler);
 				return;
@@ -1695,6 +1789,9 @@ handle_tcp_writing(netio_type *netio,
 #ifdef ECONNRESET
 			if(verbosity >= 2 || errno != ECONNRESET)
 #endif /* ECONNRESET */
+#ifdef EPIPE
+					if(verbosity >= 2 || errno != EPIPE)
+#endif /* EPIPE 'broken pipe' */
 			log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 			cleanup_tcp_handler(netio, handler);
 			return;
@@ -1725,7 +1822,7 @@ handle_tcp_writing(netio_type *netio,
 			q->tcplen = buffer_remaining(q->packet);
 			data->bytes_transmitted = 0;
 			/* Reset timeout.  */
-			handler->timeout->tv_sec = TCP_TIMEOUT;
+			handler->timeout->tv_sec = data->nsd->tcp_timeout;
 			handler->timeout->tv_nsec = 0;
 			timespec_add(handler->timeout, netio_current_time(netio));
 
@@ -1741,9 +1838,15 @@ handle_tcp_writing(netio_type *netio,
 	 * Done sending, wait for the next request to arrive on the
 	 * TCP socket by installing the TCP read handler.
 	 */
+	if (data->nsd->tcp_query_count > 0 &&
+		data->query_count >= data->nsd->tcp_query_count) {
+
+		(void) shutdown(handler->fd, SHUT_WR);
+	}
+
 	data->bytes_transmitted = 0;
 
-	handler->timeout->tv_sec = TCP_TIMEOUT;
+	handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	handler->timeout->tv_nsec = 0;
 	timespec_add(handler->timeout, netio_current_time(netio));
 
@@ -1787,9 +1890,21 @@ handle_tcp_accept(netio_type *netio,
 	addrlen = sizeof(addr);
 	s = accept(handler->fd, (struct sockaddr *) &addr, &addrlen);
 	if (s == -1) {
-		/* EINTR is a signal interrupt. The others are various OS ways
-		   of saying that the client has closed the connection. */
-		if (	errno != EINTR
+		/**
+		 * EMFILE and ENFILE is a signal that the limit of open
+		 * file descriptors has been reached. Pause accept().
+		 * EINTR is a signal interrupt. The others are various OS ways
+		 * of saying that the client has closed the connection.
+		 */
+		if (errno == EMFILE || errno == ENFILE) {
+			if (!slowaccept) {
+				slowaccept_timeout.tv_sec = NETIO_SLOW_ACCEPT_TIMEOUT;
+				slowaccept_timeout.tv_nsec = 0L;
+				timespec_add(&slowaccept_timeout, netio_current_time(netio));
+				slowaccept = 1;
+				/* We don't want to spam the logs here */
+			}
+		} else if (errno != EINTR
 			&& errno != EWOULDBLOCK
 #ifdef ECONNABORTED
 			&& errno != ECONNABORTED
@@ -1820,6 +1935,7 @@ handle_tcp_accept(netio_type *netio,
 	tcp_data->query = query_create(tcp_region, compressed_dname_offsets,
 		compression_table_size);
 	tcp_data->nsd = data->nsd;
+	tcp_data->query_count = 0;
 
 	tcp_data->tcp_accept_handler_count = data->tcp_accept_handler_count;
 	tcp_data->tcp_accept_handlers = data->tcp_accept_handlers;
@@ -1834,7 +1950,7 @@ handle_tcp_accept(netio_type *netio,
 	tcp_handler->fd = s;
 	tcp_handler->timeout = (struct timespec *) region_alloc(
 		tcp_region, sizeof(struct timespec));
-	tcp_handler->timeout->tv_sec = TCP_TIMEOUT;
+	tcp_handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	tcp_handler->timeout->tv_nsec = 0L;
 	timespec_add(tcp_handler->timeout, netio_current_time(netio));
 
@@ -1864,7 +1980,7 @@ send_children_quit(struct nsd* nsd)
 	size_t i;
 	assert(nsd->server_kind == NSD_SERVER_MAIN && nsd->this_child == 0);
 	for (i = 0; i < nsd->child_count; ++i) {
-		if (nsd->children[i].pid > 0 && nsd->children[i].child_fd > 0) {
+		if (nsd->children[i].pid > 0 && nsd->children[i].child_fd != -1) {
 			if (write(nsd->children[i].child_fd,
 				&command,
 				sizeof(command)) == -1)
@@ -1882,6 +1998,7 @@ send_children_quit(struct nsd* nsd)
 	}
 }
 
+#ifdef BIND8_STATS
 static void
 set_children_stats(struct nsd* nsd)
 {
@@ -1893,6 +2010,7 @@ set_children_stats(struct nsd* nsd)
 		nsd->children[i].handler->event_types |= NETIO_EVENT_WRITE;
 	}
 }
+#endif /* BIND8_STATS */
 
 static void
 configure_handler_event_types(size_t count,
