@@ -1,7 +1,7 @@
 /*
  * zonec.c -- zone compiler.
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -106,7 +106,7 @@ zparser_conv_hex(region_type *region, const char *hex, size_t len)
 		while (*hex) {
 			*t = 0;
 			for (i = 16; i >= 1; i -= 15) {
-				if (isxdigit(*hex)) {
+				if (isxdigit((int)*hex)) {
 					*t += hexdigit_to_int(*hex) * i;
 				} else {
 					zc_error_prev_line(
@@ -147,7 +147,7 @@ zparser_conv_hex_length(region_type *region, const char *hex, size_t len)
 		while (*hex) {
 			*t = 0;
 			for (i = 16; i >= 1; i -= 15) {
-				if (isxdigit(*hex)) {
+				if (isxdigit((int)*hex)) {
 					*t += hexdigit_to_int(*hex) * i;
 				} else {
 					zc_error_prev_line(
@@ -244,6 +244,23 @@ zparser_conv_services(region_type *region, const char *protostr,
 	return r;
 }
 
+uint16_t *
+zparser_conv_serial(region_type *region, const char *serialstr)
+{
+	uint16_t *r = NULL;
+	uint32_t serial;
+	const char *t;
+
+	serial = strtoserial(serialstr, &t);
+	if (*t != '\0') {
+		zc_error_prev_line("serial is expected");
+	} else {
+		serial = htonl(serial);
+		r = alloc_rdata_init(region, &serial, sizeof(serial));
+	}
+	return r;
+}
+
 uint16_t *
 zparser_conv_period(region_type *region, const char *periodstr)
 {
@@ -253,7 +270,7 @@ zparser_conv_period(region_type *region, const char *periodstr)
 	const char *end;
 
 	/* Allocate required space... */
-	period = (uint32_t) strtottl(periodstr, &end);
+	period = strtottl(periodstr, &end);
 	if (*end != '\0') {
 		zc_error_prev_line("time period is expected");
 	} else {
@@ -567,14 +584,14 @@ precsize_aton (char *cp, char **endptr)
 	int exponent;
 	int mantissa;
 
-	while (isdigit(*cp))
+	while (isdigit((int)*cp))
 		mval = mval * 10 + hexdigit_to_int(*cp++);
 
 	if (*cp == '.') {	/* centimeters */
 		cp++;
-		if (isdigit(*cp)) {
+		if (isdigit((int)*cp)) {
 			cmval = hexdigit_to_int(*cp++) * 10;
-			if (isdigit(*cp)) {
+			if (isdigit((int)*cp)) {
 				cmval += hexdigit_to_int(*cp++);
 			}
 		}
@@ -628,7 +645,6 @@ zparser_conv_loc(region_type *region, char *str)
 	char *start;
 	double d;
 
-
 	for(;;) {
 		deg = min = secs = 0;
 
@@ -640,17 +656,17 @@ zparser_conv_loc(region_type *region, char *str)
 
 		if (!parse_int(str, &str, &deg, "degrees", 0, 180))
 			return NULL;
-		if (!isspace(*str)) {
+		if (!isspace((int)*str)) {
 			zc_error_prev_line("space expected after degrees");
 			return NULL;
 		}
 		++str;
 
 		/* Minutes? */
-		if (isdigit(*str)) {
+		if (isdigit((int)*str)) {
 			if (!parse_int(str, &str, &min, "minutes", 0, 60))
 				return NULL;
-			if (!isspace(*str)) {
+			if (!isspace((int)*str)) {
 				zc_error_prev_line("space expected after minutes");
 				return NULL;
 			}
@@ -658,7 +674,7 @@ zparser_conv_loc(region_type *region, char *str)
 		}
 
 		/* Seconds? */
-		if (isdigit(*str)) {
+		if (isdigit((int)*str)) {
 			start = str;
 			if (!parse_int(str, &str, &i, "seconds", 0, 60)) {
 				return NULL;
@@ -668,7 +684,7 @@ zparser_conv_loc(region_type *region, char *str)
 				return NULL;
 			}
 
-			if (!isspace(*str)) {
+			if (!isspace((int)*str)) {
 				zc_error_prev_line("space expected after seconds");
 				return NULL;
 			}
@@ -711,7 +727,7 @@ zparser_conv_loc(region_type *region, char *str)
 		if (lat != 0 && lon != 0)
 			break;
 
-		if (!isspace(*str)) {
+		if (!isspace((int)*str)) {
 			zc_error_prev_line("space expected after latitude/longitude");
 			return NULL;
 		}
@@ -724,7 +740,7 @@ zparser_conv_loc(region_type *region, char *str)
 		return NULL;
 	}
 
-	if (!isspace(*str)) {
+	if (!isspace((int)*str)) {
 		zc_error_prev_line("space expected before altitude");
 		return NULL;
 	}
@@ -738,7 +754,7 @@ zparser_conv_loc(region_type *region, char *str)
 	}
 
 	/* Meters of altitude... */
-	(void)strtol(str, &str, 10);
+	(void) strtol(str, &str, 10);
 	switch(*str) {
 	case ' ':
 	case '\0':
@@ -748,7 +764,7 @@ zparser_conv_loc(region_type *region, char *str)
 		if (!parse_int(str + 1, &str, &i, "altitude fraction", 0, 99)) {
 			return NULL;
 		}
-		if (!isspace(*str) && *str != '\0' && *str != 'm') {
+		if (!isspace((int)*str) && *str != '\0' && *str != 'm') {
 			zc_error_prev_line("altitude fraction must be a number");
 			return NULL;
 		}
@@ -757,7 +773,7 @@ zparser_conv_loc(region_type *region, char *str)
 		zc_error_prev_line("altitude must be expressed in meters");
 		return NULL;
 	}
-	if (!isspace(*str) && *str != '\0')
+	if (!isspace((int)*str) && *str != '\0')
 		++str;
 
 	if (sscanf(start, "%lf", &d) != 1) {
@@ -766,16 +782,16 @@ zparser_conv_loc(region_type *region, char *str)
 
 	alt = (uint32_t) (10000000.0 + d * 100 + 0.5);
 
-	if (!isspace(*str) && *str != '\0') {
+	if (!isspace((int)*str) && *str != '\0') {
 		zc_error_prev_line("unexpected character after altitude");
 		return NULL;
 	}
 
 	/* Now parse size, horizontal precision and vertical precision if any */
-	for(i = 1; isspace(*str) && i <= 3; i++) {
+	for(i = 1; isspace((int)*str) && i <= 3; i++) {
 		vszhpvp[i] = precsize_aton(str + 1, &str);
 
-		if (!isspace(*str) && *str != '\0') {
+		if (!isspace((int)*str) && *str != '\0') {
 			zc_error_prev_line("invalid size or precision");
 			return NULL;
 		}
@@ -895,21 +911,21 @@ zparser_conv_apl_rdata(region_type *region, char *str)
  * but to "normal" (int,long,char) types
  */
 
-int32_t
-zparser_ttl2int(const char *ttlstr)
+uint32_t
+zparser_ttl2int(const char *ttlstr, int* error)
 {
 	/* convert a ttl value to a integer
 	 * return the ttl in a int
 	 * -1 on error
 	 */
 
-	int32_t ttl;
+	uint32_t ttl;
 	const char *t;
 
 	ttl = strtottl(ttlstr, &t);
 	if (*t != 0) {
 		zc_error_prev_line("invalid TTL value: %s",ttlstr);
-		ttl = -1;
+		*error = 1;
 	}
 
 	return ttl;
@@ -919,7 +935,7 @@ zparser_ttl2int(const char *ttlstr)
 void
 zadd_rdata_wireformat(uint16_t *data)
 {
-	if (parser->current_rr.rdata_count > MAXRDATALEN) {
+	if (parser->current_rr.rdata_count >= MAXRDATALEN) {
 		zc_error_prev_line("too many rdata elements");
 	} else {
 		parser->current_rr.rdatas[parser->current_rr.rdata_count].data
@@ -928,10 +944,62 @@ zadd_rdata_wireformat(uint16_t *data)
 	}
 }
 
+/**
+ * Used for TXT RR's to grow with undefined number of strings.
+ */
+void
+zadd_rdata_txt_wireformat(uint16_t *data, int first)
+{
+	rdata_atom_type *rd;
+	
+	/* First STR in str_seq, allocate 65K in first unused rdata
+	 * else find last used rdata */
+	if (first) {
+		rd = &parser->current_rr.rdatas[parser->current_rr.rdata_count];
+		if ((rd->data = (uint16_t *) region_alloc(parser->rr_region,
+			sizeof(uint16_t) + 65535 * sizeof(uint8_t))) == NULL) {
+			zc_error_prev_line("Could not allocate memory for TXT RR");
+			return;
+		}
+		parser->current_rr.rdata_count++;
+		rd->data[0] = 0;
+	}
+	else
+		rd = &parser->current_rr.rdatas[parser->current_rr.rdata_count-1];
+	
+	if ((size_t)rd->data[0] + (size_t)data[0] > 65535) {
+		zc_error_prev_line("too large rdata element");
+		return;
+	}
+	
+	memcpy((uint8_t *)rd->data + 2 + rd->data[0], data + 1, data[0]);
+	rd->data[0] += data[0];
+}
+
+/**
+ * Clean up after last call of zadd_rdata_txt_wireformat
+ */
+void
+zadd_rdata_txt_clean_wireformat()
+{
+	uint16_t *tmp_data;
+	rdata_atom_type *rd = &parser->current_rr.rdatas[parser->current_rr.rdata_count-1];
+	if ((tmp_data = (uint16_t *) region_alloc(parser->region, 
+		rd->data[0] + 2)) != NULL) {
+		memcpy(tmp_data, rd->data, rd->data[0] + 2);
+		rd->data = tmp_data;
+	}
+	else {
+		/* We could not get memory in non-volatile region */
+		zc_error_prev_line("could not allocate memory for rdata");
+		return;
+	}
+}
+
 void
 zadd_rdata_domain(domain_type *domain)
 {
-	if (parser->current_rr.rdata_count > MAXRDATALEN) {
+	if (parser->current_rr.rdata_count >= MAXRDATALEN) {
 		zc_error_prev_line("too many rdata elements");
 	} else {
 		parser->current_rr.rdatas[parser->current_rr.rdata_count].domain
@@ -1094,7 +1162,10 @@ process_rr(void)
 
 	/* We only support IN class */
 	if (rr->klass != CLASS_IN) {
-		zc_error_prev_line("only class IN is supported");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("only class IN is supported");
+		else
+			zc_error_prev_line("only class IN is supported");
 		return 0;
 	}
 
@@ -1107,34 +1178,36 @@ process_rr(void)
 		return 0;
 	}
 
+	/* Do we have the zone already? */
+	if (!zone)
+	{
+		zone = (zone_type *) region_alloc(parser->region,
+							  sizeof(zone_type));
+		zone->apex = parser->default_apex;
+		zone->soa_rrset = NULL;
+		zone->soa_nx_rrset = NULL;
+		zone->ns_rrset = NULL;
+		zone->opts = NULL;
+		zone->is_secure = 0;
+		zone->updated = 1;
+
+		zone->next = parser->db->zones;
+		parser->db->zones = zone;
+		parser->current_zone = zone;
+	}
+
 	if (rr->type == TYPE_SOA) {
 		/*
 		 * This is a SOA record, start a new zone or continue
 		 * an existing one.
 		 */
 		if (rr->owner->is_apex) {
-			/*
-			   should be error!
-			   every zone should be passed to zonec only once.
-			*/
-			zone = namedb_find_zone(parser->db, rr->owner);
-			assert(zone);
-		} else {
-			/* new zone part */
-			zone = (zone_type *) region_alloc(parser->region,
-							  sizeof(zone_type));
+			if(zone->opts && zone->opts->request_xfr)
+				zc_warning_prev_line("this SOA record was already encountered");
+			else
+				zc_error_prev_line("this SOA record was already encountered");
+		} else if (rr->owner == parser->default_apex) {
 			zone->apex = rr->owner;
-			zone->soa_rrset = NULL;
-			zone->soa_nx_rrset = NULL;
-			zone->ns_rrset = NULL;
-			zone->opts = NULL;
-			zone->is_secure = 0;
-			zone->updated = 1;
-
-			/* insert in front of zone list */
-			zone->next = parser->db->zones;
-			parser->db->zones = zone;
-
 			rr->owner->is_apex = 1;
 		}
 
@@ -1145,7 +1218,10 @@ process_rr(void)
 	if (!dname_is_subdomain(domain_dname(rr->owner),
 				domain_dname(zone->apex)))
 	{
-		zc_error_prev_line("out of zone data");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("out of zone data");
+		else
+			zc_error_prev_line("out of zone data");
 		return 0;
 	}
 
@@ -1190,40 +1266,52 @@ process_rr(void)
 	}
 
 	if(rr->type == TYPE_DNAME && rrset->rr_count > 1) {
-		zc_error_prev_line("multiple DNAMEs at the same name");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("multiple DNAMEs at the same name");
+		else
+			zc_error_prev_line("multiple DNAMEs at the same name");
 	}
 	if(rr->type == TYPE_CNAME && rrset->rr_count > 1) {
-		zc_error_prev_line("multiple CNAMEs at the same name");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("multiple CNAMEs at the same name");
+		else
+			zc_error_prev_line("multiple CNAMEs at the same name");
 	}
 	if((rr->type == TYPE_DNAME && domain_find_rrset(rr->owner, zone, TYPE_CNAME))
 	 ||(rr->type == TYPE_CNAME && domain_find_rrset(rr->owner, zone, TYPE_DNAME))) {
-		zc_error_prev_line("DNAME and CNAME at the same name");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("DNAME and CNAME at the same name");
+		else
+			zc_error_prev_line("DNAME and CNAME at the same name");
 	}
 	if(domain_find_rrset(rr->owner, zone, TYPE_CNAME) &&
 		domain_find_non_cname_rrset(rr->owner, zone)) {
-		zc_error_prev_line("CNAME and other data at the same name");
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("CNAME and other data at the same name");
+		else
+			zc_error_prev_line("CNAME and other data at the same name");
 	}
 
-#ifdef DNSSEC
-	if (rr->type == TYPE_RRSIG && rr_rrsig_type_covered(rr) == TYPE_SOA) {
+	if (rr->type == TYPE_RRSIG && rr_rrsig_type_covered(rr) == TYPE_DNSKEY) {
 		rrset->zone->is_secure = 1;
 	}
-#endif
 
 	/* Check we have SOA */
-	/* [XXX] this is dead code */
 	if (zone->soa_rrset == NULL) {
-		if (rr->type != TYPE_SOA) {
-			zc_error_prev_line(
-				"missing SOA record on top of the zone");
-		} else if (rr->owner != zone->apex) {
-			zc_error_prev_line(
-				"SOA record with invalid domain name");
-		} else {
-			zone->soa_rrset = rrset;
+		if (rr->type == TYPE_SOA) {
+			if (rr->owner != zone->apex) {
+				zc_error_prev_line(
+					"SOA record with invalid domain name");
+			} else {
+				zone->soa_rrset = rrset;
+			}
 		}
-	} else if (rr->type == TYPE_SOA) {
-		zc_error_prev_line("duplicate SOA record discarded");
+	}
+	else if (rr->type == TYPE_SOA) {
+		if(zone->opts && zone->opts->request_xfr)
+			zc_warning_prev_line("duplicate SOA record discarded");
+		else
+			zc_error_prev_line("duplicate SOA record discarded");
 		--rrset->rr_count;
 	}
 
@@ -1248,10 +1336,10 @@ domain_find_rrset_any(domain_type *domain, uint16_t type)
 	while (result) {
 		if (rrset_rrtype(result) == type) {
 			return result;
-                }
+		}
 		result = result->next;
-        }
-        return NULL;
+	}
+	return NULL;
 }
 
 /*
@@ -1525,12 +1613,14 @@ main (int argc, char **argv)
 	check_dname(db);
 
 #ifndef NDEBUG
-	fprintf(stdout, "global_region: ");
-	region_dump_stats(global_region, stderr);
-	fprintf(stdout, "\n");
-	fprintf(stdout, "db->region: ");
-	region_dump_stats(db->region, stderr);
-	fprintf(stdout, "\n");
+	if (vflag > 0) {
+		fprintf(stdout, "global_region: ");
+		region_dump_stats(global_region, stdout);
+		fprintf(stdout, "\n");
+		fprintf(stdout, "db->region: ");
+		region_dump_stats(db->region, stdout);
+		fprintf(stdout, "\n");
+	}
 #endif /* NDEBUG */
 
 	/* Close the database */
@@ -1542,8 +1632,12 @@ main (int argc, char **argv)
 
 	/* Print the total number of errors */
 	if (vflag > 0 || totalerrors > 0) {
-		fprintf(stderr, "\nzonec: done with %ld errors.\n",
+		if (totalerrors > 0) {
+			fprintf(stderr, "\nzonec: done with %ld errors.\n",
 			totalerrors);
+		} else {
+			fprintf(stdout, "\nzonec: done with no errors.\n");
+		}
 	}
 
 	/* Disable this to save some time.  */
